# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from fabric.orchestrator.swagger_server.models.base_model_ import Model
from fabric.orchestrator.swagger_server import util


class Success(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """
    def __init__(self, value: object=None):  # noqa: E501
        """Success - a model defined in Swagger

        :param value: The value of this Success.  # noqa: E501
        :type value: object
        """
        self.swagger_types = {
            'value': object
        }

        self.attribute_map = {
            'value': 'value'
        }
        self._value = value

    @classmethod
    def from_dict(cls, dikt) -> 'Success':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Success of this Success.  # noqa: E501
        :rtype: Success
        """
        return util.deserialize_model(dikt, cls)

    @property
    def value(self) -> object:
        """Gets the value of this Success.

        JSON object  # noqa: E501

        :return: The value of this Success.
        :rtype: object
        """
        return self._value

    @value.setter
    def value(self, value: object):
        """Sets the value of this Success.

        JSON object  # noqa: E501

        :param value: The value of this Success.
        :type value: object
        """

        self._value = value
